/*
 * Copyright (c) 2006
 *
 * written by Simon Barrick
 * simon@fourthwall.ndo.co.uk
 *
 * PointSlide SOP
 */
#define MAXFLT FLT_MAX

#include <UT/UT_DSOVersion.h>
#include <UT/UT_Math.h>
#include <PRM/PRM_Include.h>
#include <OP/OP_Operator.h>
#include <OP/OP_OperatorTable.h>
#include <UT/UT_Interrupt.h>
#include <SOP/SOP_Node.h>

#include <GB/GB_Edge.h>
#include <GB/GB_PointRef.h>
#include <GB/GB_EdgeGroup.h>
#include <GEO/GEO_Point.h>
#include <GEO/GEO_Vertex.h>
#include <GEO/GEO_Closure.h>
#include <GEO/GEO_PointTree.h>

#include "SOP_PointSlide.h"


void
newSopOperator(OP_OperatorTable *table)
{
     table->addOperator
     (
       new OP_Operator
       (  "pointslide",
          "Point Slide",
           SOP_PointSlide::myConstructor,
           SOP_PointSlide::myTemplateList,
           1,                             // Min Inputs
           1,                             // Max Inputs
           SOP_PointSlide::myVariables    // Local Variables
        )
      );
}

static PRM_Name        names[] = {
    PRM_Name("type",  "Type"),
    PRM_Name("dist",   "Dist"),
    PRM_Name("iterations", "Iterations"),
    PRM_Name("reverse", "Reverse"),
    PRM_Name("split", "Split Points"),
    PRM_Name("radius", "Soft Radius"),
    PRM_Name("clamp", "Clamp split points"),
};

static PRM_Name typeMenuNames[] =
{
  PRM_Name( "abs", "Absolute" ),

  PRM_Name( "uv", "Parametric" ),

  PRM_Name(0)

};

static PRM_ChoiceList   typeMenu((PRM_ChoiceListType)(PRM_CHOICELIST_EXCLUSIVE
                                                    | PRM_CHOICELIST_REPLACE),
                                                      typeMenuNames);

PRM_Template
SOP_PointSlide::myTemplateList[] = {
  //group
    PRM_Template(PRM_STRING, 1, &PRMgroupName, 0, &SOP_Node::pointGroupMenu),
  //type
    PRM_Template(PRM_ORD, 1, &names[0], PRMzeroDefaults, &typeMenu),
  //dist
    PRM_Template(PRM_FLT_J, 1, &names[1], PRMzeroDefaults, 0, &PRMfullunituiRange),
  //dist
    PRM_Template(PRM_INT_J, 1, &names[2], PRMoneDefaults, 0, &PRMdivisionRange),
  //split
    PRM_Template(PRM_TOGGLE, 1, &names[4], PRMzeroDefaults),
  //reverse
    PRM_Template(PRM_TOGGLE, 1, &names[3], PRMzeroDefaults),
  //soft radius
    PRM_Template(PRM_FLT_J, 1, &names[5], PRMzeroDefaults, 0, &PRMnonNegativeRange),
  //clamp
    PRM_Template(PRM_TOGGLE, 1, &names[6], PRMzeroDefaults),

    PRM_Template(),
};

// First, lets get a bunch of uniqe integers
enum {
        VAR_PTS,
        VAR_PTE,
        VAR_E,
     };

// Next, we define our local variable table
CH_LocalVariable
SOP_PointSlide::myVariables[] = {
      { "PT0", VAR_PTS, 0 },
      { "PT1", VAR_PTE, 0 },
      { "EDGE", VAR_E, 0 },
      { 0, 0, 0 },
};

float
SOP_PointSlide::getVariableValue(int index, int) {

    if (myCurrPoint < 0) // Sorry, we're in an invalid state
    return 0;

    switch (index)
    {
      case VAR_PTS: return myCurrPoint;
      case VAR_PTE: return myCurrEndPoint;
      case VAR_E:   return myCurrEdge;
    }
    return 0; //Whoops, forgot a variable
}

OP_Node *
SOP_PointSlide::myConstructor(OP_Network *net, const char *name, OP_Operator *op)
{
    return new SOP_PointSlide(net, name, op);
}

SOP_PointSlide::SOP_PointSlide(OP_Network *net, const char *name, OP_Operator *op)
  : SOP_Node(net, name, op)
{
  myCurrPoint = -1;
}

SOP_PointSlide::~SOP_PointSlide() {}

const char *
SOP_PointSlide::inputLabel( unsigned int input) const
{
 switch (input) {
   case 0:
     return "Input Geometry";
     break;
   default:
     return "WTF?";
     break;
 }
}

unsigned
SOP_PointSlide::disableParms()
{
    unsigned changed = 0;

    changed += enableParm("clamp", SPLIT());

    return changed;
}

OP_ERROR
SOP_PointSlide::cookInputGroups(OP_Context &context, int alone)
{
    // If we are called by the handle, then "alone" equals 1.  In that
    // case, we have to lock the inputs oursevles, and unlock them
    // before exiting this method.
    if (alone) if (lockInputs(context) >= UT_ERROR_ABORT) return error();

    UT_String  grp_name;

    // The "gdp" variable is only available if we are called from the SOP
    // itself.  So, if we are called by a handle, we have to get the
    // geometry oursevles.
    GU_Detail *pgdp = alone ? (GU_Detail *)inputGeo(0, context) : gdp;

    myGroup = 0;

    getGroups(grp_name);    // Get the group string.

    // If the group string is not null, then we try to parse the group.
    if (grp_name.isstring())
    {
      edgegroup = true;

      myGroup = parseEdgeGroups((const char *)grp_name, pgdp);

      // If the group is not valid, then the group string is invalid
      // as well.  Thus, we add an error to this SOP.
      if (!myGroup)
      {
          addError(SOP_ERR_BADGROUP, grp_name);
      }
      else if (!alone)
      {
          // If the parsed group is valid, then we want to highlight
          // only the group.  The second argument of "1" means that
          // we want the selection to have the same type as our group.
          select(*(GB_BaseGroup *)myGroup, 1);
      }
    }
    else edgegroup = false;
    // This is where we notify our handles (if any) if the inputs have changed.
    //checkInputChanged(0, -1, myDetailGroupPair, pgdp, myGroup);

    // If we are called by the handles, then we have to unlock our inputs.
    if (alone)
    {
      destroyAdhocGroups();
      unlockInputs();
    }

    return error();
}

int
SOP_PointSlide::splitVertices(GU_Detail *gdp)
{
  GEO_Point         *ppt,*ppt0,*ppt1;
  GB_Edge           *edge;
  GB_PointRefArray  prefarr(gdp,0);
  UT_IntArray       primlist0,primlist1;
  GEO_Vertex        *vtx;
  GEO_Primitive     *prim;
  int               i,j,test;
  UT_PtrArray<GEO_Vertex *> vtxArray;

  GB_EdgeGroup *myGroupCopy;
  myGroupCopy = createAdhocEdgeGroup(*gdp, "tmpedgegroup");

  FOR_ALL_GROUP_EDGES(myEdgeGroup, edge)
    test = myGroupCopy->add(edge->p0(),edge->p1(),edge->prim());

  GEO_Closure connectedPrims(*gdp);

  int newpointcount = 0;

  //loop through all edges in edge group
  FOR_ALL_GROUP_EDGES(myEdgeGroup, edge)
  {
    if (reverse)
    {
      ppt0 = (GEO_Point*)(edge->p1());
      ppt1 = (GEO_Point*)(edge->p0());
    }
    else
    {
      ppt0 = (GEO_Point*)(edge->p0());
      ppt1 = (GEO_Point*)(edge->p1());
    }

    bool foundsplit = true;
    int count = 0;
    while (foundsplit)
    {

      foundsplit = false;

      //if all prims using point also use current edge
      //break
      connectedPrims.findPolysUsingEdge(*ppt0, *ppt1, primlist0, prefarr);
      connectedPrims.findPrimsUsingPoint(ppt0, primlist1, prefarr);

      if (!primlist0.entries() || primlist0.entries()==primlist1.entries())
      {
        test = myGroupCopy->remove(edge);
        break;
      }

      for(i=0; i<primlist0.entries(); i++)
      {
        prim = gdp->primitives()(primlist0[i]);

      //for each vertex in prims check if it uses edge ppt0
        for(j=0; j<prim->getVertexCount(); j++)
        {
          vtx = &prim->getVertex(j);
          ppt = vtx->getPt();

          //if it does add to vtxArray
          if (ppt->getNum() == ppt0->getNum())
          {
            vtxArray(vtxArray.append()) = vtx;
            foundsplit = true;
          }
        }
      }

      if (foundsplit)
      {
        //append point into the gdp
        ppt = gdp->appendPoint(ppt0->getPos(), ppt0->getAttrib(), 0);
        newpointcount++;

        for(j = 0;j < vtxArray.entries(); j++)
        {
          vtx = vtxArray[j];
          //insert point into prim
          vtx->setPt(ppt);
        }
        //update edge group with new points
        test = myGroupCopy->remove(edge);
        test = myGroupCopy->add(ppt, ppt1, prim);;
      }
      vtxArray.entries(0);
    }
  }

  myEdgeGroup->clearEntries();
  FOR_ALL_GROUP_EDGES(myGroupCopy, edge)
      test = myEdgeGroup->add(edge->p0(),edge->p1(),edge->prim());

  test = destroyAdhocGroup(myGroupCopy);

  select(*(GB_BaseGroup *)myEdgeGroup, 1);

  return newpointcount;
}

OP_ERROR
SOP_PointSlide::cookMySop(OP_Context &context)
{
  GB_Edge       *edge;
  GEO_Point     *ppt, *ppt0, *ppt1;
  int           test,i,j;
  float         *Natt;
  UT_PtrArray<GEO_Point*> pointlist;
  const GB_PointGroup   *pointgroup=NULL;

  float now = context.myTime;

  // Before we do anything, we must lock our inputs.  Before returning,
  // we have to make sure that the inputs get unlocked.
  if( lockInputs( context ) >= UT_ERROR_ABORT ) return error();

  //copy input geomtry as we aren't creating any
  duplicateSource( 0, context );

  // Here we determine which groups we have to work on.  We only
  //  handle point groups.
  if (error() < UT_ERROR_ABORT && cookInputGroups(context) < UT_ERROR_ABORT)
  {
    //for each pair of points in the group the first is the point to move
    //the second point determines the direction
    int edgeCount = 0;

    int iterations = ITER(now);
    reverse = REV();
    bool split = SPLIT();
    float radius = RADIUS(now);

    GU_Detail *softgdp;
    softgdp = new GU_Detail();

    if (edgegroup)
    {
      GEO_PointTree pointtree;

      myEdgeGroup = createAdhocEdgeGroup(*gdp, "edgegroup");

      FOR_ALL_GROUP_EDGES(myGroup, edge)
        test = myEdgeGroup->add(edge->p0(),edge->p1(),edge->prim());

      UT_BitArray   bitArray(gdp->points().entries());
      bitArray.setAllBits(1);

      //if clamp split
      if (CLAMP())
        FOR_ALL_GROUP_EDGES(myEdgeGroup, edge)
        {
          if (reverse)
            ppt = (GEO_Point*)(edge->p1());
          else
            ppt = (GEO_Point*)(edge->p0());
          bitArray.setBit(ppt->getNum(),0);
        }

      if (split)
      {
        //test we only have polys and exit gracefully if not
        GEO_Closure closure(*gdp);
        GB_PointGroup *tmpPointGroup;
        tmpPointGroup = createAdhocPointGroup(*gdp, "tmppointgroup");
        tmpPointGroup = closure.getPointClosure(*myEdgeGroup);

        GB_PrimitiveGroup *tmpPrimGroup;
        tmpPrimGroup = createAdhocPrimGroup(*gdp, "tmpprimgroup");
        tmpPrimGroup = closure.getPrimitiveClosure(*tmpPointGroup);

        GEO_Primitive *myprim;
        //only test prims that have edges in myGroup
        //primclosure...
        FOR_ALL_GROUP_PRIMITIVES(gdp, tmpPrimGroup, myprim)
        {
          if (myprim->getPrimitiveId() != 1)
          {
            addError(SOP_ERR_INVALID_SRC , "I can only split polygons, sorry");
            destroyAdhocGroups();
            unlockInputs();
            return error();
          }
        }

        test = destroyAdhocGroup(tmpPointGroup);
        test = destroyAdhocGroup(tmpPrimGroup);

        int newpointcount = splitVertices(gdp);
      //expand bitArray by number of new points, all set to zero
        for(i=0; i<newpointcount; i++)
          bitArray.append(0);
      }


      for(i=0; i<iterations; i++)
      {
        if (i)
          softgdp->clearAndDestroy();
        int Noff = softgdp->addPointAttrib("N", sizeof(float)*3, GB_ATTRIB_FLOAT,0);

        FOR_ALL_GROUP_EDGES(myEdgeGroup, edge)
        {
          if (reverse && !split)
          {
            ppt0 = (GEO_Point*)(edge->p1());
            ppt1 = (GEO_Point*)(edge->p0());
          }
          else
          {
            ppt0 = (GEO_Point*)(edge->p0());
            ppt1 = (GEO_Point*)(edge->p1());
          }

          UT_Vector3 dir =  ppt1->getPos() - ppt0->getPos();

          if (!TYPE()) dir.normalize();

          myCurrPoint = ppt0->getNum();
          myCurrEndPoint = ppt1->getNum();
          myCurrEdge = edgeCount;

          dir *= DIST(now)/iterations;

          ppt0->getPos()(0) += dir.x();
          ppt0->getPos()(1) += dir.y();
          ppt0->getPos()(2) += dir.z();

          edgeCount++;

          //if soft radius > 0
          if (radius > 0 )
          {
            //add pt0 to a softgdp
            ppt = softgdp->appendPoint();
            ppt->getPos() = ppt0->getPos();
            //add edge as normal to softgdp
            Natt = (float*)ppt->getAttribData(Noff);
            Natt[0] = dir.x();
            Natt[1] = dir.y();
            Natt[2] = dir.z();
            if (!i)
              bitArray.setBit(ppt0->getNum(),0);
          }
        }
        if (radius > 0)
        {
          pointtree.build(softgdp,pointgroup);
          //for all points not in myEdgeGroup
          FOR_ALL_GPOINTS(gdp, ppt)
          {
            if (bitArray.getBit(ppt->getNum()))
            {
              //find the all the points in softgdp within the soft radius
              test = pointtree.findAllClosePt(ppt->getPos(), radius, pointlist);

              UT_Vector3 totalDir(0,0,0);
              for(j=0;j<pointlist.entries();j++)
              {
                UT_Vector3 testP = pointlist[j]->getPos();
                Natt= (float*)pointlist[j]->getAttribData(Noff);
                UT_Vector3 dir(Natt[0],Natt[1],Natt[2]);

                //do a weighted sum of all the directions and distances
                float weight = (1-SYSsmooth(0.0f,radius,(testP-ppt->getPos()).length()));
                totalDir += weight*dir;
              }
              if (pointlist.entries())
              {
                totalDir /= pointlist.entries();
                //offset the points by the amount
                ppt->getPos()(0) += totalDir.x();
                ppt->getPos()(1) += totalDir.y();
                ppt->getPos()(2) += totalDir.z();
              }
            }
          }
        }
      }

      int test = destroyAdhocGroup(myEdgeGroup);

    }

    delete softgdp;
  }

  myCurrPoint = -1;

  unlockInputs();

  return error();
}
